@@ -1,9 +1,10 @@
 <?php
 
-/*
+/**
  +-----------------------------------------------------------------------+
  | This file is part of the Roundcube Webmail client                     |
- | Copyright (C) 2005-2012, The Roundcube Dev Team                       |
+ |                                                                       |
+ | Copyright (C) The Roundcube Dev Team                                  |
  |                                                                       |
  | Licensed under the GNU General Public License version 3 or            |
  | any later version with exceptions for skins & plugins.                |
@@ -25,15 +26,19 @@
  */
 class rcube_db
 {
+    /** @var string Database implementation (mysql, postgres, etc.) */
     public $db_provider;
 
     protected $db_dsnw;               // DSN for write operations
     protected $db_dsnr;               // DSN for read operations
+    protected $db_dsnw_array;         // DSN for write operations
+    protected $db_dsnr_array;         // DSN for read operations
     protected $db_connected = false;  // Already connected ?
     protected $db_mode;               // Connection mode
+    protected $db_pconn = false;      // Persistent connections flag
     protected $dbh;                   // Connection handle
-    protected $dbhs = array();
-    protected $table_connections = array();
+    protected $dbhs = [];
+    protected $table_connections = [];
 
     protected $db_error     = false;
     protected $db_error_msg = '';
@@ -43,15 +48,24 @@ class rcube_db
     protected $tables;
     protected $variables;
 
-    protected $options = array(
+    protected $options = [
         // column/table quotes
         'identifier_start' => '"',
         'identifier_end'   => '"',
-    );
+        // date/time input format
+        'datetime_format'  => 'Y-m-d H:i:s',
+        'table_prefix'     => '',
+    ];
 
     const DEBUG_LINE_LENGTH = 4096;
     const DEFAULT_QUOTE     = '`';
 
+    const TYPE_SQL    = 'sql';
+    const TYPE_INT    = 'integer';
+    const TYPE_BOOL   = 'bool';
+    const TYPE_STRING = 'string';
+
+
     /**
      * Factory, returns driver-specific instance of the class
      *
@@ -64,23 +78,26 @@ class rcube_db
     public static function factory($db_dsnw, $db_dsnr = '', $pconn = false)
     {
         $driver     = strtolower(substr($db_dsnw, 0, strpos($db_dsnw, ':')));
-        $driver_map = array(
+        $driver_map = [
             'sqlite2' => 'sqlite',
             'sybase'  => 'mssql',
             'dblib'   => 'mssql',
             'mysqli'  => 'mysql',
             'oci'     => 'oracle',
             'oci8'    => 'oracle',
-        );
+        ];
 
         $driver = isset($driver_map[$driver]) ? $driver_map[$driver] : $driver;
         $class  = "rcube_db_$driver";
 
         if (!$driver || !class_exists($class)) {
-            rcube::raise_error(array('code' => 600, 'type' => 'db',
-                'line' => __LINE__, 'file' => __FILE__,
-                'message' => "Configuration error. Unsupported database driver: $driver"),
-                true, true);
+            rcube::raise_error([
+                    'code' => 600, 'type' => 'db',
+                    'line' => __LINE__, 'file' => __FILE__,
+                    'message' => "Configuration error. Unsupported database driver: $driver"
+                ],
+                true, true
+            );
         }
 
         return new $class($db_dsnw, $db_dsnr, $pconn);
@@ -110,7 +127,7 @@ class rcube_db
 
         $this->options['table_prefix']  = $config->get('db_prefix');
         $this->options['dsnw_noread']   = $config->get('db_dsnw_noread', false);
-        $this->options['table_dsn_map'] = array_map(array($this, 'table_name'), $config->get('db_table_dsn', array()));
+        $this->options['table_dsn_map'] = array_map([$this, 'table_name'], $config->get('db_table_dsn', []));
     }
 
     /**
@@ -125,15 +142,15 @@ class rcube_db
         $this->db_error_msg = null;
 
         // return existing handle
-        if ($this->dbhs[$mode]) {
-            $this->dbh = $this->dbhs[$mode];
+        if (!empty($this->dbhs[$mode])) {
+            $this->dbh     = $this->dbhs[$mode];
             $this->db_mode = $mode;
+
             return $this->dbh;
         }
 
         // connect to database
         if ($dbh = $this->conn_create($dsn)) {
-            $this->dbh          = $dbh;
             $this->dbhs[$mode]  = $dbh;
             $this->db_mode      = $mode;
             $this->db_connected = true;
@@ -149,10 +166,6 @@ class rcube_db
         $dsn_string  = $this->dsn_string($dsn);
         $dsn_options = $this->dsn_options($dsn);
 
-        if ($this->db_pconn) {
-            $dsn_options[PDO::ATTR_PERSISTENT] = true;
-        }
-
         // Connect
         try {
             // with this check we skip fatal error on PDO object creation
@@ -162,25 +175,32 @@ class rcube_db
 
             $this->conn_prepare($dsn);
 
-            $dbh = new PDO($dsn_string, $dsn['username'], $dsn['password'], $dsn_options);
+            $username = isset($dsn['username']) ? $dsn['username'] : null;
+            $password = isset($dsn['password']) ? $dsn['password'] : null;
+
+            $this->dbh = new PDO($dsn_string, $username, $password, $dsn_options);
 
             // don't throw exceptions or warnings
-            $dbh->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_SILENT);
+            $this->dbh->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_SILENT);
 
-            $this->conn_configure($dsn, $dbh);
+            $this->conn_configure($dsn, $this->dbh);
         }
         catch (Exception $e) {
             $this->db_error     = true;
             $this->db_error_msg = $e->getMessage();
 
-            rcube::raise_error(array('code' => 500, 'type' => 'db',
-                'line' => __LINE__, 'file' => __FILE__,
-                'message' => $this->db_error_msg), true, false);
+            rcube::raise_error([
+                    'code' => 500, 'type' => 'db',
+                    'line' => __LINE__, 'file' => __FILE__,
+                    'message' => $this->db_error_msg
+                ],
+                true, false
+            );
 
             return null;
         }
 
-        return $dbh;
+        return $this->dbh;
     }
 
     /**
@@ -205,8 +225,8 @@ class rcube_db
     /**
      * Connect to appropriate database depending on the operation
      *
-     * @param string $mode Connection mode (r|w)
-     * @param boolean $force Enforce using the given mode
+     * @param string $mode  Connection mode (r|w)
+     * @param bool   $force Enforce using the given mode
      */
     public function db_connect($mode, $force = false)
     {
@@ -223,7 +243,7 @@ class rcube_db
         // Already connected
         if ($this->db_connected) {
             // connected to db with the same or "higher" mode (if allowed)
-            if ($this->db_mode == $mode || $this->db_mode == 'w' && !$force && !$this->options['dsnw_noread']) {
+            if ($this->db_mode == $mode || $this->db_mode == 'w' && !$force && empty($this->options['dsnw_noread'])) {
                 return;
             }
         }
@@ -262,14 +282,14 @@ class rcube_db
                 $table = $m[2];
 
                 // always use direct mapping
-                if ($this->options['table_dsn_map'][$table]) {
+                if (!empty($this->options['table_dsn_map'][$table])) {
                     $mode = $this->options['table_dsn_map'][$table];
                     break;  // primary table rules
                 }
                 else if ($mode == 'r') {
                     // connected to db with the same or "higher" mode for this table
                     $db_mode = $this->table_connections[$table];
-                    if ($db_mode == 'w' && !$this->options['dsnw_noread']) {
+                    if ($db_mode == 'w' && empty($this->options['dsnw_noread'])) {
                         $mode = $db_mode;
                     }
                 }
@@ -286,7 +306,7 @@ class rcube_db
     /**
      * Activate/deactivate debug mode
      *
-     * @param boolean $dbg True if SQL queries should be logged
+     * @param bool $dbg True if SQL queries should be logged
      */
     public function set_debug($dbg = true)
     {
@@ -300,12 +320,13 @@ class rcube_db
      */
     protected function debug($query)
     {
-        if ($this->options['debug_mode']) {
+        if (!empty($this->options['debug_mode'])) {
             if (($len = strlen($query)) > self::DEBUG_LINE_LENGTH) {
                 $diff  = $len - self::DEBUG_LINE_LENGTH;
                 $query = substr($query, 0, self::DEBUG_LINE_LENGTH)
                     . "... [truncated $diff bytes]";
             }
+
             rcube::write_log('sql', '[' . (++$this->db_index) . '] ' . $query . ';');
         }
     }
@@ -329,7 +350,7 @@ class rcube_db
     /**
      * Connection state checker
      *
-     * @return boolean True if in connected state
+     * @return bool True if in connected state
      */
     public function is_connected()
     {
@@ -366,12 +387,12 @@ class rcube_db
      * @param string SQL query to execute
      * @param mixed  Values to be inserted in query
      *
-     * @return number  Query handle identifier
+     * @return PDOStatement|false  Query handle or False on error
      */
     public function query()
     {
         $params = func_get_args();
-        $query = array_shift($params);
+        $query  = array_shift($params);
 
         // Support one argument of type array, instead of n arguments
         if (count($params) == 1 && is_array($params[0])) {
@@ -389,7 +410,7 @@ class rcube_db
      * @param int    Number of rows for LIMIT statement
      * @param mixed  Values to be inserted in query
      *
-     * @return PDOStatement|bool Query handle or False on error
+     * @return PDOStatement|false Query handle or False on error
      */
     public function limitquery()
     {
@@ -409,7 +430,7 @@ class rcube_db
      * @param int    $numrows Number of rows for LIMIT statement
      * @param array  $params  Values to be inserted in query
      *
-     * @return PDOStatement|bool Query handle or False on error
+     * @return PDOStatement|false Query handle or False on error
      */
     protected function _query($query, $offset, $numrows, $params)
     {
@@ -436,7 +457,7 @@ class rcube_db
 
         if (count($params)) {
             while ($pos = strpos($query, '?', $pos)) {
-                if ($query[$pos+1] == '?') {  // skip escaped '?'
+                if (isset($query[$pos+1]) && $query[$pos+1] == '?') {  // skip escaped '?'
                     $pos += 2;
                 }
                 else {
@@ -452,8 +473,8 @@ class rcube_db
 
         // replace escaped '?' and quotes back to normal, see self::quote()
         $query = str_replace(
-            array('??', self::DEFAULT_QUOTE.self::DEFAULT_QUOTE),
-            array('?', self::DEFAULT_QUOTE),
+            ['??', self::DEFAULT_QUOTE.self::DEFAULT_QUOTE],
+            ['?', self::DEFAULT_QUOTE],
             $query
         );
 
@@ -503,7 +524,7 @@ class rcube_db
         $in  = false;
 
         while ($pos = strpos($query, $quote, $pos)) {
-            if ($query[$pos+1] == $quote) {  // skip escaped quote
+            if (isset($query[$pos+1]) && $query[$pos+1] == $quote) {  // skip escaped quote
                 $pos += 2;
             }
             else {
@@ -526,28 +547,71 @@ class rcube_db
 
     /**
      * Helper method to handle DB errors.
-     * This by default logs the error but could be overriden by a driver implementation
+     * This by default logs the error but could be overridden by a driver implementation
+     *
+     * @param string $query Query that triggered the error
      *
-     * @param string Query that triggered the error
      * @return mixed Result to be stored and returned
      */
     protected function handle_error($query)
     {
         $error = $this->dbh->errorInfo();
 
-        if (empty($this->options['ignore_key_errors']) || !in_array($error[0], array('23000', '23505'))) {
+        if (empty($this->options['ignore_key_errors']) || !in_array($error[0], ['23000', '23505'])) {
             $this->db_error = true;
             $this->db_error_msg = sprintf('[%s] %s', $error[1], $error[2]);
 
-            rcube::raise_error(array('code' => 500, 'type' => 'db',
-                'line' => __LINE__, 'file' => __FILE__,
-                'message' => $this->db_error_msg . " (SQL Query: $query)"
-                ), true, false);
+            if (empty($this->options['ignore_errors'])) {
+                rcube::raise_error([
+                        'code' => 500, 'type' => 'db', 'line' => __LINE__, 'file' => __FILE__,
+                        'message' => $this->db_error_msg . " (SQL Query: $query)"
+                    ], true, false);
+            }
         }
 
         return false;
     }
 
+    /**
+     * INSERT ... ON DUPLICATE KEY UPDATE (or equivalent).
+     * When not supported by the engine we do UPDATE and INSERT.
+     *
+     * @param string $table   Table name (should be already passed via table_name() with quoting)
+     * @param array  $keys    Hash array (column => value) of the unique constraint
+     * @param array  $columns List of columns to update
+     * @param array  $values  List of values to update (number of elements
+     *                        should be the same as in $columns)
+     *
+     * @return PDOStatement|bool Query handle or False on error
+     * @todo Multi-insert support
+     */
+    public function insert_or_update($table, $keys, $columns, $values)
+    {
+        $columns = array_map(function($i) { return "`$i`"; }, $columns);
+        $sets    = array_map(function($i) { return "$i = ?"; }, $columns);
+        $where   = $keys;
+
+        array_walk($where, function(&$val, $key) {
+            $val = $this->quote_identifier($key) . " = " . $this->quote($val);
+        });
+
+        // First try UPDATE
+        $result = $this->query("UPDATE $table SET " . implode(", ", $sets)
+            . " WHERE " . implode(" AND ", $where), $values);
+
+        // if UPDATE fails use INSERT
+        if ($result && !$this->affected_rows($result)) {
+            $cols  = implode(', ', array_map(function($i) { return "`$i`"; }, array_keys($keys)));
+            $cols .= ', ' . implode(', ', $columns);
+            $vals  = implode(', ', array_map(function($i) { return $this->quote($i); }, $keys));
+            $vals .= ', ' . rtrim(str_repeat('?, ', count($columns)), ', ');
+
+            $result = $this->query("INSERT INTO $table ($cols) VALUES ($vals)", $values);
+        }
+
+        return $result;
+    }
+
     /**
      * Get number of affected rows for the last query
      *
@@ -571,7 +635,8 @@ class rcube_db
      * If no query handle is specified, the last query will be taken as reference
      *
      * @param mixed $result Optional query handle
-     * @return mixed   Number of rows or false on failure
+     *
+     * @return mixed Number of rows or false on failure
      * @deprecated This method shows very poor performance and should be avoided.
      */
     public function num_rows($result = null)
@@ -610,9 +675,7 @@ class rcube_db
             $table = $this->table_name($table);
         }
 
-        $id = $this->dbh->lastInsertId($table);
-
-        return $id;
+        return $this->dbh->lastInsertId($table);
     }
 
     /**
@@ -621,7 +684,7 @@ class rcube_db
      *
      * @param mixed $result Optional query handle
      *
-     * @return mixed Array with col values or false on failure
+     * @return array|false Array with col values or false on failure
      */
     public function fetch_assoc($result = null)
     {
@@ -634,7 +697,7 @@ class rcube_db
      *
      * @param mixed $result Optional query handle
      *
-     * @return mixed Array with col values or false on failure
+     * @return array|false Array with col values or false on failure
      */
     public function fetch_array($result = null)
     {
@@ -647,7 +710,7 @@ class rcube_db
      * @param mixed $result Optional query handle
      * @param int   $mode   Fetch mode identifier
      *
-     * @return mixed Array with col values or false on failure
+     * @return array|false Array with col values or false on failure
      */
     protected function _fetch_row($result, $mode)
     {
@@ -693,9 +756,10 @@ class rcube_db
         if ($this->tables === null) {
             $q = $this->query("SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES"
                 . " WHERE TABLE_TYPE = 'BASE TABLE'"
-                . " ORDER BY TABLE_NAME");
+                . " ORDER BY TABLE_NAME"
+            );
 
-            $this->tables = $q ? $q->fetchAll(PDO::FETCH_COLUMN, 0) : array();
+            $this->tables = $q ? $q->fetchAll(PDO::FETCH_COLUMN, 0) : [];
         }
 
         return $this->tables;
@@ -710,14 +774,13 @@ class rcube_db
      */
     public function list_cols($table)
     {
-        $q = $this->query('SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = ?',
-            array($table));
+        $q = $this->query('SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = ?', $table);
 
         if ($q) {
             return $q->fetchAll(PDO::FETCH_COLUMN, 0);
         }
 
-        return array();
+        return [];
     }
 
     /**
@@ -777,6 +840,30 @@ class rcube_db
         return $this->last_result = $this->dbh->rollBack();
     }
 
+    /**
+     * Release resources related to the last query result.
+     * When we know we don't need to access the last query result we can destroy it
+     * and release memory. Useful especially if the query returned big chunk of data.
+     */
+    public function reset()
+    {
+        $this->last_result = null;
+    }
+
+    /**
+     * Terminate database connection.
+     */
+    public function closeConnection()
+    {
+        $this->db_connected = false;
+        $this->db_index     = 0;
+
+        // release statement and connection resources
+        $this->last_result  = null;
+        $this->dbh          = null;
+        $this->dbhs         = [];
+    }
+
     /**
      * Formats input so it can be safely used in a query
      *
@@ -787,6 +874,10 @@ class rcube_db
      */
     public function quote($input, $type = null)
     {
+        if ($input instanceof rcube_db_param) {
+            return (string) $input;
+        }
+
         // handle int directly for better performance
         if ($type == 'integer' || $type == 'int') {
             return intval($input);
@@ -796,6 +887,10 @@ class rcube_db
             return 'NULL';
         }
 
+        if ($input instanceof DateTime) {
+            return $this->quote($input->format($this->options['datetime_format']));
+        }
+
         if ($type == 'ident') {
             return $this->quote_identifier($input);
         }
@@ -806,16 +901,16 @@ class rcube_db
         }
 
         if ($this->dbh) {
-            $map = array(
+            $map = [
                 'bool'    => PDO::PARAM_BOOL,
                 'integer' => PDO::PARAM_INT,
-            );
+            ];
 
             $type = isset($map[$type]) ? $map[$type] : PDO::PARAM_STR;
 
             return strtr($this->dbh->quote($input, $type),
                 // escape ? and `
-                array('?' => '??', self::DEFAULT_QUOTE => self::DEFAULT_QUOTE.self::DEFAULT_QUOTE)
+                ['?' => '??', self::DEFAULT_QUOTE => self::DEFAULT_QUOTE.self::DEFAULT_QUOTE]
             );
         }
 
@@ -877,14 +972,25 @@ class rcube_db
     {
         $start = $this->options['identifier_start'];
         $end   = $this->options['identifier_end'];
-        $name  = array();
+        $name  = [];
 
         foreach (explode('.', $str) as $elem) {
-            $elem = str_replace(array($start, $end), '', $elem);
+            $elem = str_replace([$start, $end], '', $elem);
             $name[] = $start . $elem . $end;
         }
 
-        return implode($name, '.');
+        return implode('.', $name);
+    }
+
+    /**
+     * Create query parameter object
+     *
+     * @param mixed  $value Parameter value
+     * @param string $type  Parameter type (one of rcube_db::TYPE_* constants)
+     */
+    public function param($value, $type = null)
+    {
+        return new rcube_db_param($this, $value, $type);
     }
 
     /**
@@ -896,13 +1002,15 @@ class rcube_db
      */
     public function now($interval = 0)
     {
+        $result = 'now()';
+
         if ($interval) {
-            $add = ' ' . ($interval > 0 ? '+' : '-') . ' INTERVAL ';
-            $add .= $interval > 0 ? intval($interval) : intval($interval) * -1;
-            $add .= ' SECOND';
+            $result .= ' ' . ($interval > 0 ? '+' : '-') . ' INTERVAL '
+                . ($interval > 0 ? intval($interval) : intval($interval) * -1)
+                . ' SECOND';
         }
 
-        return "now()" . $add;
+        return $result;
     }
 
     /**
@@ -934,7 +1042,7 @@ class rcube_db
      *
      * @param string $field Field name
      *
-     * @return string  SQL statement to use in query
+     * @return string SQL statement to use in query
      * @deprecated
      */
     public function unixtimestamp($field)
@@ -948,10 +1056,11 @@ class rcube_db
      * @param int $timestamp Unix timestamp
      *
      * @return string Date string in db-specific format
+     * @deprecated
      */
     public function fromunixtime($timestamp)
     {
-        return date("'Y-m-d H:i:s'", $timestamp);
+        return $this->quote(date($this->options['datetime_format'], $timestamp));
     }
 
     /**
@@ -975,11 +1084,11 @@ class rcube_db
     public function concat(/* col1, col2, ... */)
     {
         $args = func_get_args();
-        if (is_array($args[0])) {
+        if (!empty($args) && is_array($args[0])) {
             $args = $args[0];
         }
 
-        return '(' . join(' || ', $args) . ')';
+        return '(' . implode(' || ', $args) . ')';
     }
 
     /**
@@ -1002,6 +1111,7 @@ class rcube_db
             foreach (get_object_vars($input) as $idx => $value) {
                 $input->$idx = self::encode($value);
             }
+
             return $input;
         }
         else if (is_array($input)) {
@@ -1028,7 +1138,7 @@ class rcube_db
         // use Base64 encoding to workaround issues with invalid
         // or null characters in serialized string (#1489142)
         if ($serialized) {
-            // Keep backward compatybility where base64 wasn't used
+            // Keep backward compatibility where base64 wasn't used
             if (strpos(substr($input, 0, 16), ':') !== false) {
                 return self::decode(@unserialize($input));
             }
@@ -1040,12 +1150,14 @@ class rcube_db
             foreach (get_object_vars($input) as $idx => $value) {
                 $input->$idx = self::decode($value);
             }
+
             return $input;
         }
         else if (is_array($input)) {
             foreach ($input as $idx => $value) {
                 $input[$idx] = self::decode($value);
             }
+
             return $input;
         }
 
@@ -1063,7 +1175,7 @@ class rcube_db
     public function table_name($table, $quoted = false)
     {
         // let plugins alter the table name (#1489837)
-        $plugin = rcube::get_instance()->plugins->exec_hook('db_table_name', array('table' => $table));
+        $plugin = rcube::get_instance()->plugins->exec_hook('db_table_name', ['table' => $table]);
         $table = $plugin['table'];
 
         // add prefix to the table name if configured
@@ -1092,8 +1204,8 @@ class rcube_db
     /**
      * Set DSN connection to be used for the given table
      *
-     * @param string Table name
-     * @param string DSN connection ('r' or 'w') to be used
+     * @param string $table Table name
+     * @param string $mode  DSN connection ('r' or 'w') to be used
      */
     public function set_table_dsn($table, $mode)
     {
@@ -1103,7 +1215,7 @@ class rcube_db
     /**
      * MDB2 DSN string parser
      *
-     * @param string $sequence Secuence name
+     * @param string $sequence Sequence name
      *
      * @return array DSN parameters
      */
@@ -1153,7 +1265,6 @@ class rcube_db
         }
 
         // Find protocol and hostspec
-
         // $dsn => proto(proto_opts)/database
         if (preg_match('|^([^(]+)\((.*?)\)/?(.*?)$|', $dsn, $match)) {
             $proto       = $match[1];
@@ -1169,9 +1280,9 @@ class rcube_db
                 && strpos($dsn, '/', 2) !== false
                 && $parsed['phptype'] == 'oci8'
             ) {
-                //oracle's "Easy Connect" syntax:
-                //"username/password@[//]host[:port][/service_name]"
-                //e.g. "scott/tiger@//mymachine:1521/oracle"
+                // Oracle's "Easy Connect" syntax:
+                // "username/password@[//]host[:port][/service_name]"
+                // e.g. "scott/tiger@//mymachine:1521/oracle"
                 $proto_opts = $dsn;
                 $pos = strrpos($proto_opts, '/');
                 $dsn = substr($proto_opts, $pos + 1);
@@ -1187,34 +1298,35 @@ class rcube_db
         }
 
         // process the different protocol options
-        $parsed['protocol'] = (!empty($proto)) ? $proto : 'tcp';
+        $parsed['protocol'] = !empty($proto) ? $proto : 'tcp';
         $proto_opts = rawurldecode($proto_opts);
         if (strpos($proto_opts, ':') !== false) {
             list($proto_opts, $parsed['port']) = explode(':', $proto_opts);
         }
-        if ($parsed['protocol'] == 'tcp') {
+        if ($parsed['protocol'] == 'tcp' && strlen($proto_opts)) {
             $parsed['hostspec'] = $proto_opts;
         }
         else if ($parsed['protocol'] == 'unix') {
             $parsed['socket'] = $proto_opts;
         }
 
-        // Get dabase if any
+        // Get database if any
         // $dsn => database
         if ($dsn) {
             // /database
             if (($pos = strpos($dsn, '?')) === false) {
                 $parsed['database'] = rawurldecode($dsn);
-            // /database?param1=value1&param2=value2
             }
             else {
+                // /database?param1=value1&param2=value2
                 $parsed['database'] = rawurldecode(substr($dsn, 0, $pos));
                 $dsn = substr($dsn, $pos + 1);
+
                 if (strpos($dsn, '&') !== false) {
                     $opts = explode('&', $dsn);
                 }
                 else { // database?param1=value1
-                    $opts = array($dsn);
+                    $opts = [$dsn];
                 }
                 foreach ($opts as $opt) {
                     list($key, $value) = explode('=', $opt);
@@ -1224,6 +1336,19 @@ class rcube_db
                     }
                 }
             }
+
+            // remove problematic suffix (#7034)
+            $parsed['database'] = preg_replace('/;.*$/', '', $parsed['database']);
+
+            // Resolve relative path to the sqlite database file
+            // so for example it works with Roundcube Installer
+            if (!empty($parsed['phptype']) && !empty($parsed['database'])
+                && stripos($parsed['phptype'], 'sqlite') === 0
+                && $parsed['database'][0] != '/'
+                && strpos($parsed['database'], ':') === false
+            ) {
+                $parsed['database'] = INSTALL_PATH . $parsed['database'];
+            }
         }
 
         return $parsed;
@@ -1238,18 +1363,18 @@ class rcube_db
      */
     protected function dsn_string($dsn)
     {
-        $params = array();
+        $params = [];
         $result = $dsn['phptype'] . ':';
 
-        if ($dsn['hostspec']) {
+        if (isset($dsn['hostspec'])) {
             $params[] = 'host=' . $dsn['hostspec'];
         }
 
-        if ($dsn['port']) {
+        if (isset($dsn['port'])) {
             $params[] = 'port=' . $dsn['port'];
         }
 
-        if ($dsn['database']) {
+        if (isset($dsn['database'])) {
             $params[] = 'dbname=' . $dsn['database'];
         }
 
@@ -1269,7 +1394,19 @@ class rcube_db
      */
     protected function dsn_options($dsn)
     {
-        $result = array();
+        $result = [];
+
+        if ($this->db_pconn) {
+            $result[PDO::ATTR_PERSISTENT] = true;
+        }
+
+        if (!empty($dsn['prefetch'])) {
+            $result[PDO::ATTR_PREFETCH] = (int) $dsn['prefetch'];
+        }
+
+        if (!empty($dsn['timeout'])) {
+            $result[PDO::ATTR_TIMEOUT] = (int) $dsn['timeout'];
+        }
 
         return $result;
     }
@@ -1277,27 +1414,40 @@ class rcube_db
     /**
      * Execute the given SQL script
      *
-     * @param string SQL queries to execute
+     * @param string $sql SQL queries to execute
      *
-     * @return boolen True on success, False on error
+     * @return boolean True on success, False on error
      */
     public function exec_script($sql)
     {
         $sql  = $this->fix_table_names($sql);
         $buff = '';
+        $exec = '';
 
         foreach (explode("\n", $sql) as $line) {
-            if (preg_match('/^--/', $line) || trim($line) == '')
+            $trimmed = trim($line);
+            if ($trimmed == '' || preg_match('/^--/', $trimmed)) {
                 continue;
+            }
 
-            $buff .= $line . "\n";
-            if (preg_match('/(;|^GO)$/', trim($line))) {
-                $this->query($buff);
+            if ($trimmed == 'GO') {
+                $exec = $buff;
+            }
+            else if ($trimmed[strlen($trimmed)-1] == ';') {
+                $exec = $buff . substr(rtrim($line), 0, -1);
+            }
+
+            if ($exec) {
+                $this->query($exec);
                 $buff = '';
+                $exec = '';
                 if ($this->db_error) {
                     break;
                 }
             }
+            else {
+                $buff .= $line . "\n";
+            }
         }
 
         return !$this->db_error;
@@ -1313,10 +1463,10 @@ class rcube_db
         }
 
         $sql = preg_replace_callback(
-            '/((TABLE|TRUNCATE|(?<!ON )UPDATE|INSERT INTO|FROM'
-            . '| ON(?! (DELETE|UPDATE))|REFERENCES|CONSTRAINT|FOREIGN KEY|INDEX)'
+            '/((TABLE|TRUNCATE( TABLE)?|(?<!ON )UPDATE|INSERT INTO|FROM'
+            . '| ON(?! (DELETE|UPDATE))|REFERENCES|CONSTRAINT|FOREIGN KEY|INDEX|UNIQUE( INDEX)?)'
             . '\s+(IF (NOT )?EXISTS )?[`"]*)([^`"\( \r\n]+)/',
-            array($this, 'fix_table_names_callback'),
+            [$this, 'fix_table_names_callback'],
             $sql
         );
 
@@ -1328,6 +1478,24 @@ class rcube_db
      */
     protected function fix_table_names_callback($matches)
     {
-        return $matches[1] . $this->options['table_prefix'] . $matches[count($matches)-1];
+        $prefix = $this->options['table_prefix'];
+
+        // Schema prefix (ends with a dot)
+        if ($prefix[strlen($prefix)-1] === '.') {
+            // These can't have a schema prefix
+            if (preg_match('/(CONSTRAINT|UNIQUE|INDEX)[\s\t`"]*$/', $matches[1])) {
+                $prefix = '';
+            }
+            else {
+                // check if the identifier is quoted, then quote the prefix
+                $last = $matches[1][strlen($matches[1])-1];
+
+                if ($last === '`' || $last === '"') {
+                    $prefix = substr($prefix, 0, -1) . $last . '.' . $last;
+                }
+            }
+        }
+
+        return $matches[1] . $prefix . $matches[count($matches)-1];
     }
 }
